"""Helper functions."""

def generate_attribute_string(self, attributes):
    output = ""
    for attribute in attributes:
        output += f"{attribute}: {getattr(self, attribute)}\n"
    return output



import asyncio
import ipaddress
import logging
import ssl
import socket
import pathlib

import aiohttp

from .device import Device
from .execution import Execution
from .hue import Hue
from .hdmi import Hdmi
from .errors import raise_error, RequestError, Unauthorized
from .hsb_cacert import HSB_CACERT

MIN_API_LEVEL = 4

logger = logging.getLogger(__name__)

class CommonNameInserterResolver(aiohttp.DefaultResolver):

    def __init__(self, common_name, loop=None, *args, **kwargs):
        super().__init__(loop=loop, *args, **kwargs)
        self._common_name = common_name

    async def resolve(self, host, port=0, family=socket.AF_INET):
        hosts = []
        if host.startswith('_'):
            # Host was an IP address that was mangled to force a DNS lookup (_ are not valid)
            # and with that forced lookup end up in this call.
            # This is needed as IP addresses don't need lookup,
            # but I need to set the hostname to the common_name for certificate validation
            # and this seems to be the only place I could hook into

            # Generate a suitable entry in the hosts list
            hosts.append({
                'host': host[1:],
                'port': port,
                'family': family,
                'proto': 6, # TCP I think
                'flags': socket.AI_NUMERICHOST,
            })
        else:
            hosts.append(await super().resolve(host, port=port, family=family))

        for host in hosts:
            host['hostname'] = self._common_name

        return hosts


class HueSyncBox:
    """Control a Philips Hue Play HDMI Sync Box."""

    def __init__(self, host, id, access_token=None, port=443, path='/api'):
        self._host = host
        self._id = id
        self._access_token = access_token
        self._port = port
        self._path = path

        self._clientsession = self._get_clientsession()

        # API endpoints
        self.device = None
        self.execution = None
        self.hue = None
        self.hdmi = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    def _get_clientsession(self):
        """
        Get a clientsession that is tuned for communication with the Hue Syncbox
        """
        context = ssl.create_default_context(cadata=HSB_CACERT)
        context.hostname_checks_common_name = True

        connector = aiohttp.TCPConnector(
            enable_cleanup_closed=True, # Home Assistant sets it so lets do it also
            ssl=context,
            limit_per_host=1, # Syncbox can handle a limited amount of connections, only take what we need
            resolver=CommonNameInserterResolver(self._id) # Use custom resolver to get certificate validation on common_name working
        )

        return aiohttp.ClientSession(connector=connector)

    @property
    def access_token(self):
        return self._access_token

    async def is_registered(self):
        try:
            await self.request('get', '/registrations')
            return True
        except Unauthorized:
            return False
        return False

    async def register(self, application_name, instance_name, use_registered_token=True):
        """
        Register with the huesyncbox

        application_name : Userfriendly name of your application
        instance_name : The specific instance of your application, e.g. a specific device the application is running on
        use_registered_token: When true use the token (if obtained) for subsequent requests
        """
        response = await self.request('post', '/registrations', {
            "appName": application_name,
            "instanceName": instance_name
        }, auth=False) # Make sure to _not_ use a possibly invalid token as it will be rejected

        info = None
        if response:
            info = {
                'registration_id': response['registrationId'],
                'access_token': response['accessToken']
            }

            if use_registered_token:
                self._access_token = info['access_token']

        return info

    async def unregister(self, registration_id):
        """Unregister application from the huesyncbox, you can only unregister the id associated with the token in use."""
        await self.request('delete', f'/registrations/{registration_id}')

    async def initialize(self):
        await self.update()
        if self.device.api_level < MIN_API_LEVEL:
            logger.error("This library requires at least API version %s. Please update the Philips Hue Play HDMI Sync Box.", MIN_API_LEVEL)

    async def close(self):
        await self._clientsession.close()

    async def update(self):
        response = await self.request('get', '')
        if response:
            self.device = Device(response['device'], self.request)
            self.execution = Execution(response['execution'], self.request)
            self.hue = Hue(response['hue'], self.request)
            self.hdmi = Hdmi(response['hdmi'], self.request)

    def _mangled_host(self):
        """
        Returns the hostname or a modified hostname in case the host is an IP address
        to make sure DNS lookups are required as that allows to use the common_name
        instead of servername for certificate validation.
        """
        try:
            ipaddress.ip_address(self._host)
            return f"_{self._host}"
        except ValueError:
            pass
        return self._host

    async def request(self, method, path, data=None, auth=True):
        """Make a request to the API."""

        if self._clientsession.closed:
            # Avoid runtime errors when connection is closed.
            # This solves an issue when Updates were scheduled and HA was shutdown
            return None

        url = f'https://{self._mangled_host()}:{self._port}{self._path}/v1{path}'

        try:
            headers = {'Content-Type': 'application/json'}
            if auth and self._access_token:
                headers['Authorization'] = f'Bearer {self._access_token}'
            logger.debug('%s, %s, %s' % (method, url, data))
            async with self._clientsession.request(method, url, json=data, headers=headers) as resp:
                logger.debug('%s, %s' % (resp.status, await resp.text('utf-8')))

                data = None
                if resp.content_type == 'application/json':
                    data = await resp.json()
                    if resp.status != 200:
                        _raise_on_error(data)
                return data
        except aiohttp.client_exceptions.ClientError as err:
            raise RequestError(
                'Error requesting data from {}: {}'.format(self._host, err)
            ) from None


def _raise_on_error(data):
    """Check response for error message."""
    raise_error(data['code'], data["message"])



from .helpers import generate_attribute_string

class Input():

    def __init__(self, id, raw):
        self.id = id
        self._raw = raw

    @property
    def name(self):
        """Friendly name of the input."""
        return self._raw['name']

    @property
    def type(self):
        """Type of the input."""
        return self._raw['type']

    @property
    def status(self):
        """Status of the input: unplugged, plugged, linked, unknown"""
        return self._raw['status']

    @property
    def last_sync_mode(self):
        """Last sync mode of the input."""
        return self._raw['lastSyncMode']


class Hdmi:
    """Represent Hdmi config of huesyncbox."""

    def __init__(self, raw, request):
        self._raw = raw
        self._request = request
        self._update_inputs_and_output()

    def __str__(self):
        attributes = ["content_specs", "video_sync_supported", "audio_sync_supported", "inputs", "output"]
        return generate_attribute_string(self, attributes)

    def __eq__(self, other: object) -> bool:
        return self._raw == other._raw

    def _update_inputs_and_output(self):
        inputs = []
        for key, value in self._raw.items():
            if key.startswith('input'):
                inputs.append(Input(key, value))
        self._inputs = inputs
        self._output = Input('output', self._raw['output'])

    @property
    def content_specs(self):
        """Content specs of current input of huesyncbox."""
        return self._raw['contentSpecs']

    @property
    def video_sync_supported(self):
        """Indicates if syncing is supported on video content."""
        return self._raw['videoSyncSupported']

    @property
    def audio_sync_supported(self):
        """Indicates if syncing is supported on audio content."""
        return self._raw['audioSyncSupported']

    @property
    def inputs(self):
        """HDMI inputs of the huesyncbox."""
        return self._inputs

    @property
    def output(self):
        """HDMI output of the huesyncbox."""
        return self._output

    async def update(self):
        response = await self._request('get', '/hdmi')
        if response:
            self._raw = response
            self._update_inputs_and_output()


from .helpers import generate_attribute_string

class Device():
    """Represent Device config."""

    def __init__(self, raw, request):
        self._raw = raw
        self._request = request

    def __str__(self):
        attributes = ["name", "device_type", "unique_id", "ip_address", "api_level", "firmware_version"]
        return generate_attribute_string(self, attributes)

    def __eq__(self, other: object) -> bool:
        return self._raw == other._raw

    @property
    def name(self):
        """Friendly name of the device."""
        return self._raw['name']

    @property
    def device_type(self):
        """Device Type identifier."""
        return self._raw['deviceType']

    @property
    def unique_id(self):
        """Capitalized hex string of the 6 byte / 12 characters device id without delimiters. Used as unique id on label, certificate common name, hostname etc."""
        return self._raw['uniqueId']

    @property
    def ip_address(self):
        """Local IP address of the device."""
        return self._raw['ipAddress']

    @property
    def api_level(self):
        """Supported API level of the device."""
        return self._raw['apiLevel']

    @property
    def firmware_version(self):
        """User readable version of the device firmware, starting with decimal major .minor .maintenance format e.g. 1.12.3."""
        return self._raw['firmwareVersion']

    async def update(self):
        response = await self._request('get', '/device')
        if response:
            self._raw = response


from .helpers import generate_attribute_string

class Group:
    """Represent a group on the Hue bridge"""

    def __init__(self, id, raw):
        self._id = id
        self._raw = raw

    @property
    def id(self):
        """Id of the group."""
        return self._id

    @property
    def name(self):
        """Friendly name of the entertainment group."""
        return self._raw['name']

    @property
    def num_lights(self):
        """Number of lights in the entertainment group."""
        return self._raw['numLights']

    @property
    def active(self):
        """Indicates if the group is actively streaming (either from Syncbox or other source)."""
        return self._raw['active']

    @property
    def owner(self):
        """
        User friendly name of the application that is streaming on the associated bridge.
        Only exposed if active is true
        """
        return self._raw['owner'] if 'owner' in self._raw else None


class Hue:
    """Represent Hue config."""

    def __init__(self, raw, request):
        self._raw = raw
        self._request = request
        self._groups = Hue._build_groups(self._raw)

    def __str__(self):
        attributes = ["bridge_unique_id", "bridge_ip_address", "connection_state", "groups"]
        return generate_attribute_string(self, attributes)

    def __eq__(self, other: object) -> bool:
        return self._raw == other._raw

    @staticmethod
    def _build_groups(raw):
        groups = []
        for key, value in raw['groups'].items():
            groups.append(Group(key, value))
        return groups

    @property
    def bridge_unique_id(self):
        """16 character ascii hex string bridge identifier."""
        return self._raw['bridgeUniqueId']

    @property
    def bridge_ip_address(self):
        """Readable, dot IPv4 address of the paired bridge EG 192.168.1.50."""
        return self._raw['bridgeIpAddress']

    @property
    def connection_state(self):
        """uninitialized, disconnected, connecting, unauthorized, connected, invalidgroup, streaming"""
        return self._raw['connectionState']

    @property
    def groups(self):
        """
        All available entertainment areas on the current bridge.
        When this object is not available, it means the bridge groups have not been retrieved yet.
        When the object is empty, it means there are no entertainment areas on the bridge.
        When the bridge connection is lost, the last known values are remembered.
        Determining whether values may be outdated can be done based on connectionState.
        """
        return self._groups

    async def set_group_active(self, id, active):
        data = { 'active': active }
        await self._request('put', f'hue/groups/{id}', data=data)

    async def update(self):
        response = await self._request('get', '/hue')
        if response:
            self._raw = response
            self._groups = Hue._build_groups(self._raw)
